from pygame_ui_toolkit.presets import button_size_change
from pygame_ui_toolkit.presets import button_colour_change
from pygame_ui_toolkit.elements import input
from pygame_ui_toolkit import utils
from pygame_ui_toolkit import pygame


stored_variables = {}


def assign_variables(text_input: input.TextInput) -> None:
    """Assign variables that need to be used later to a global dict with the input object as the key."""

    current_variables = {
        "on_select" : text_input.on_selected,
        "on_deselect" : text_input.on_deselect
    }

    stored_variables[text_input] = current_variables


def change_colour(text_input: input.TextInput, normal_event_func: callable, new_event_func: callable) -> None:
    """Change the colour of the text input button object."""
    button_object = text_input.input_button.button_object

    utils.call_func(normal_event_func, text_input)
    new_event_func(button_object)


def on_selected_func(text_input: input.TextInput) -> None:
    """The function called when the input is selected."""
    normal = stored_variables[text_input]["on_select"]
    new = button_colour_change.on_click_func

    change_colour(text_input, normal, new)


def on_deselect_func(text_input: input.TextInput) -> None:
    """The function called when the input is deselected."""
    normal = stored_variables[text_input]["on_deselect"]
    new = button_colour_change.on_normal_func

    change_colour(text_input, normal, new)


def create_text_input(deselected_colour: tuple[int], selected_colour: tuple[int], normal_size: tuple[int] | int, hover_size: tuple[int] | int, click_size: tuple[int] | int, surface: pygame.Surface, x: int, y: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, on_selected: callable = None, on_deselect: callable = None, on_text_input: callable = None, corner_radius: int = -1, text: str = "", prefix_text: str = "", min_font_size: int = 10, change_font_size: bool = True, antialias: bool = False) -> input.RectTextInput:
    """
    Return a new RectTextInput that will automatically change colour when selcted and change size when clicked or hovered.

    Ensure to call the input's update() method each frame.

    Notes
    -----
    normal_size, hover_size and click_size should be provided as a tuple in the form (width, height)
    - e.g. a value of (100, 50) would result in a text input with width 100 and height 50
    """

    text_input = input.RectTextInput(surface, x, y, deselected_colour, normal_size[0], normal_size[1], font_colour, font_size, font_name, on_click, on_hover, on_normal, on_selected, on_deselect, on_text_input, corner_radius, False, text, prefix_text, min_font_size, change_font_size, antialias)

    change_existing_text_input(text_input, deselected_colour, selected_colour, normal_size, hover_size, click_size)

    return text_input


def change_existing_text_input(text_input: input.TextInput, deselected_colour: tuple[int], selected_colour: tuple[int], normal_size: tuple[int] | int, hover_size: tuple[int] | int, click_size: tuple[int] | int) -> None:
    """
    Update a RectTextInput that will automatically change colour when selcted and change size when clicked or hovered.

    PolygonTextInput and BorderedPolygonTextInput types are not supported.
    
    Notes
    -----
    - if text_input is a RectTextInput or BorderedRectTextInput then normal_size, hover_size and click_size should be a tuple in the form (width, height)
    - if text_input is a CircleTextInput or BorderedCircleTextInput then normal_size, hover_size and click_size should be a single int corresponding to the radius
    """
    
    assign_variables(text_input)
    
    button_object = text_input.input_button.button_object

    text_input.on_selected = on_selected_func
    text_input.on_deselect = on_deselect_func

    button_size_change.change_existing_button(button_object, normal_size, hover_size, click_size)
    button_colour_change.assign_variables(button_object, deselected_colour, deselected_colour, selected_colour, None, None, None)


from pygame_ui_toolkit import pygame
from pygame_ui_toolkit import utils


class TextBox:
    """
    The base text box class that all other text boxes inherit from.

    This class should not be used directly.
    Instead use PolygonTextBox, RectTextBox, CircleTextBox, BorderedPolygonTextBox, BorderedRectTextBox or BorderedCircleTextBox.

    Attributes
    ----------
    surface : pygame.Surface
        the surface that the text box is drawn to
    text : str
        the text displayed on the text box
    background_colour : tuple[int]
        the colour of the text box
    font_colour : tuple[int]
        the colour of the text
    font : pygame.font.Font
        the font object used to blit text
    antialias : bool, optional
        whether the text is drawn with antialias (defaults to False)
    text_surface : pygame.Surface
        the surface used to render text
    text_rect : pygame.Rect
        the rect position that the text is drawn to

    Methods
    -------
    get_text()
        return a text surface to draw and a rect object to draw it to
    update_text()
        change the text, font colour, font size or font name of the displayed text
    blit_text()
        draw the text to the screen
    """

    def __init__(self, surface: pygame.Surface, text: str, background_colour: tuple[int], font_colour: tuple[int], font_size: int, font_name: str | None = None, antialias: bool = False) -> None:
        """Construct the necessary attributes of the TextBox object."""
        self.surface = surface
        
        self.text = text
        
        self.background_colour = background_colour
        self.font_colour = font_colour

        self.font = pygame.font.Font(font_name, font_size)

        self.antialias = antialias

        self.text_surface, self.text_rect = self.get_text()

    def get_text(self) -> tuple[pygame.Surface, pygame.Rect]:
        """Return a text surface to draw and a rect object to draw it to."""
        text_surface = self.font.render(self.text, self.antialias, self.font_colour)
        
        text_rect = text_surface.get_rect()

        if hasattr(self, "center"):
            text_rect.center = self.center
        else:
            raise Exception("TextBox class is not supposed to be used directly. Use another class like RectTextBox instead")

        return text_surface, text_rect

    def update_text(self, new_text: str, new_font_colour: tuple[int], new_font_size: int, new_font_name: str | None) -> None:
        """Change the text, font colour, font size or font name of the displayed text."""
        utils.update_font_attrs(self, new_text, new_font_colour, new_font_name, new_font_size)

        self.text_surface, self.text_rect = self.get_text()

    def blit_text(self) -> None:
        """Draw the text to the screen."""
        self.surface.blit(self.text_surface, self.text_rect)


class PolygonTextBox(TextBox):
    """
    A text box in the shape of any polygon.

    This inherits from TextBox.

    Attributes
    ----------
    This class contains all attributes from the TextBox class, alongside these additional ones:
    points : list[tuple[int]]
        a list of the coordinates of each point on the polygon
    center : tuple[int]
        the coordinates of the center of the text box shape

    Methods
    -------
    This class contains all methods from the TextBox class, alongside these additional ones:
    get_center()
        return the center of the polygon by averaging all points
    draw() - overwritten from TextBox
        draw the text and text box to the screen
    """

    def __init__(self, surface: pygame.Surface, points: list[tuple[int]], text: str, background_colour: tuple[int], font_colour: tuple[int], font_size: int, font_name: str | None = None, antialias: bool = False) -> None:
        """Construct the necessary attributes for the PolygonTextBox object."""
        self.points = points
        self.center = self.get_center()

        super().__init__(surface, text, background_colour, font_colour, font_size, font_name, antialias)

    def get_center(self) -> tuple[int]:
        """Return the center of the polygon by averaging all points."""
        x_values = [i[0] for i in self.points]
        y_values = [i[1] for i in self.points]

        x = sum(x_values) // len(x_values)
        y = sum(y_values) // len(y_values)

        return x, y
    
    def draw(self) -> None:
        """Draw the text and text box to the screen."""
        pygame.draw.polygon(self.surface, self.background_colour, self.points)
        self.blit_text()


class RectTextBox(TextBox):
    """
    A text box in the shape of a rectangle.

    This inherits from TextBox.

    Attributes
    ----------
    This class contains all attributes from the TextBox class, alongside these additional ones:
    x : int
        the x coordinate of the center of the rectangle
    y : int
        the y coordinate of the center of the rectangle
    center : tuple[int]
        the coordinates of the center of the text box shape
    width : int
        the width (parallel to the x axis) of the rectangle
    height : int
        the height (parallel to the y axis) of the rectangle
    corner_radius : int, optional
        the radius of the rounded corners of the rectange, where -1 means no rounded corners (defaults to -1)

    Methods
    -------
    This class contains all methods from the TextBox class, alongside these additional ones:
    draw() - overwritten from TextBox
        draw the text and text box to the screen
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, width: int, height: int, text: str, background_colour: tuple[int], font_colour: tuple[int], font_size: int, font_name: str | None = None, corner_radius: int = -1, antialias: bool = False) -> None:        
        """Construct the necessary attributes for the RectTextBox object."""
        self.x = x
        self.y = y
        
        self.center = (x, y)

        self.width = width
        self.height = height

        self.corner_radius = corner_radius

        super().__init__(surface, text, background_colour, font_colour, font_size, font_name, antialias)
        
    def draw(self) -> None:
        """Draw the text and text box to the screen."""
        pygame.draw.rect(self.surface, self.background_colour, (self.x - self.width // 2, self.y - self.height // 2, self.width, self.height))

        self.blit_text()


class CircleTextBox(TextBox):
    """
    A text box in the shape of a circle.

    This inherits from TextBox.

    Attributes
    ----------
    This class contains all attributes from the TextBox class, alongside these additional ones:
    x : int
        the x coordinate of the center of the rectangle
    y : int
        the y coordinate of the center of the rectangle
    center : tuple[int]
        the coordinates of the center of the text box shape
    radius : int
        the radius of the circle

    Methods
    -------
    This class contains all methods from the TextBox class, alongside these additional ones:
    draw() - overwritten from TextBox
        draw the text and text box to the screen
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, radius: int, text: str, background_colour: tuple[int], font_colour: tuple[int], font_size: int, font_name: str | None = None, antialias: bool = False) -> None:
        """Construct the necessary attributes for the CircleTextBox object."""
        self.x = x
        self.y = y

        self.center = (x, y)

        self.radius = radius
        
        super().__init__(surface, text, background_colour, font_colour, font_size, font_name, antialias)
    
    def draw(self) -> None:
        """Draw the text and text box to the screen."""
        pygame.draw.circle(self.surface, self.background_colour, self.center, self.radius)
        self.blit_text()


class BorderedPolygonTextBox(PolygonTextBox):
    """
    A text box in the shape of any polygon with a border.

    This inherits from PolygonTextBox.

    Attributes
    ----------
    This class contains all attributes from the PolygonTextBox class, alongside these additional ones:
    border_colour : tuple[int]
        the colour of the drawn border
    border_width : int
        the width of the drawn border

    Methods
    -------
    This class contains all methods from the PolygonTextBox class, alongside these overwritten ones:
    draw() - overwritten from PolygonTextBox
        draw the text and text box with border to the screen
    """

    def __init__(self, surface: pygame.Surface, points: list[tuple[int, int]], text: str, background_colour: tuple[int], border_colour: tuple[int], border_width: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, antialias: bool = False) -> None:
        """Construct the necessary attributes for the BorderedPolygonTextBox object."""
        super().__init__(surface, points, text, background_colour, font_colour, font_size, font_name, antialias)

        self.border_colour = border_colour
        self.border_width = border_width

    def draw(self) -> None:
        """Draw the text and text box with border to the screen."""
        pygame.draw.polygon(self.surface, self.background_colour, self.points)
        pygame.draw.polygon(self.surface, self.border_colour, self.points, self.border_width)

        self.blit_text()


class BorderedRectTextBox(RectTextBox):
    """
    A text box in the shape of a rectangle with a border.

    This inherits from RectTextBox.

    Attributes
    ----------
    This class contains all attributes from the RectTextBox class, alongside these additional ones:
    border_colour : tuple[int]
        the colour of the drawn border
    border_width : int
        the width of the drawn border

    Methods
    -------
    This class contains all methods from the RectTextBox class, alongside these overwritten ones:
    draw() - overwritten from RectTextBox
        draw the text and text box with border to the screen
    """
        
    def __init__(self, surface: pygame.Surface, x: int, y: int, width: int, height: int, text: str, background_colour: tuple[int], border_colour: tuple[int], border_width: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, antialias: bool = False) -> None:
        """Construct the necessary attributes for the BorderedRectTextBox object."""
        super().__init__(surface, x, y, width, height, text, background_colour, font_colour, font_size, font_name, antialias)

        self.border_colour = border_colour
        self.border_width = border_width

    def draw(self) -> None:
        """Draw the text and text box with border to the screen."""
        pygame.draw.rect(self.surface, self.background_colour, (self.x - self.width // 2, self.y - self.height // 2, self.width, self.height))
        pygame.draw.rect(self.surface, self.border_colour, (self.x - self.width // 2, self.y - self.height // 2, self.width, self.height), self.border_width)

        self.blit_text()


class BorderedCircleTextBox(CircleTextBox):
    """
    A text box in the shape of a circle with a border.

    This inherits from CircleTextBox.

    Attributes
    ----------
    This class contains all attributes from the CircleTextBox class, alongside these additional ones:
    border_colour : tuple[int]
        the colour of the drawn border
    border_width : int
        the width of the drawn border

    Methods
    -------
    This class contains all methods from the CircleTextBox class, alongside these overwritten ones:
    draw() - overwritten from CircleTextBox
        draw the text and text box with border to the screen
    """
        
    def __init__(self, surface: pygame.Surface, x: int, y: int, radius: int, text: str, background_colour: tuple[int], border_colour: tuple[int], border_width: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, antialias: bool = False) -> None:
        """Construct the necessary attributes for the BorderedCircleTextBox object."""
        super().__init__(surface, x, y, radius, text, background_colour, font_colour, font_size, font_name, antialias)

        self.border_colour = border_colour
        self.border_width = border_width

    def draw(self) -> None:
        """Draw the text and text box with border to the screen."""
        pygame.draw.circle(self.surface, self.background_colour, self.center, self.radius)
        pygame.draw.circle(self.surface, self.border_colour, self.center, self.radius, self.border_width)

        self.blit_text()


from pygame_ui_toolkit import pygame
from pygame_ui_toolkit import utils


class Button:
    """
    The base class that all other button classes inherit from.

    This class should not be used directly. 
    Instead use RectButton, CircleButton, PolygonButton, BorderedRectButton, BorderedCircleButton, or BorderedPolygonButton.

    Attributes
    ----------
    surface : pygame.Surface
        the surface that the button is drawn onto
    x : int
        the x position of the center of the button
    y : int
        the y position of the center of the button
    on_click : callable | None, optional
        the function that is called when the button is clicked. If the function accepts 1 argument, self is passed into it (defaults to None)
    on_hover : callable | None, optional
        the function that is called when the button is hovered over. If the function accepts 1 argument, self is passed into it (defaults to None)
    on_normal : callable | None, optional
        the function that is called when the button is neither clicked or hovered over. If the function accepts 1 argument, self is passed into it (defaults to None)
    click_once : bool, optional
        if True, the on_click and on_hover functions will be called once per click or hover event, otherwise they will be called every ferame the button is clicked or hovered (defaults to True)
    
    Methods
    -------
    update_clicked(currently_clicked: bool)
        update the clicked attribute.
    update_hovered(currently_hovered: bool)
        update the hovered attribute.
    call_func(func: callable)
        call utils.call_func(func, self).
    check_click()
        return whether the button is ciurrently being clicked.
    check_hover()
        return whether the button is being hovered over.
    update()
        draw the button and call the on_click, on_hover and on_normal functions where appropriate.
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the necessary attributes for the Button object"""
        self.surface = surface

        self.x = x
        self.y = y

        self.on_click = on_click
        self.on_hover = on_hover
        self.on_normal = on_normal

        self.click_once = click_once

        self.clicked = False
        self.hovered = False

    def update_clicked(self, currently_clicked: bool) -> bool:
        """
        Update the clicked attribute.
        
        Parameters
        ----------

        currently_clicked : bool
            whether the mouse is currently over the button and the mouse button is pressed

        Returns
        -------
        bool
            whether on_click should be called
        """
        if not self.click_once:
            self.clicked = currently_clicked
            return currently_clicked
        
        if not currently_clicked:
            self.clicked = False
            return False
        
        if self.clicked:
            return False
        else:
            self.clicked = True
            return True
        
    def update_hovered(self, currently_hovered: bool) -> bool:
        """
        Update the hovered attribute.
        
        Parameters
        ----------

        currently_hovered : bool
            whether the mouse is currently over the button

        Returns
        -------
        bool
            whether on_hover should be called
        """
        if not self.click_once:
            self.hovered = currently_hovered
            return currently_hovered
        
        if not currently_hovered:
            self.hovered = False
            return False
        
        if self.hovered:
            return False
        else:
            self.hovered = True
            return True

    get_pos = lambda self: (self.x, self.y)

    def call_func(self, func: callable) -> None:
        """Call utils.call_func(func, self)."""
        if func == None:
            return
        else:
            utils.call_func(func, self)
        
    def check_click(self) -> bool:
        """Return whether the button is ciurrently being clicked."""
        mouse_over = self.mouse_over()
        mouse_down = pygame.mouse.get_pressed(3)[0]

        click = mouse_over and mouse_down

        return self.update_clicked(click)
    
    def check_hover(self) -> bool:
        """Return whether the button is currently being hovered over."""
        mouse_over = self.mouse_over()

        return self.update_hovered(mouse_over)
    
    def update(self) -> None:
        """
        Draw the button and call the on_click, on_hover and on_normal functions where appropriate.

        This should be called once per frame.

        An exception is raised if the button object does not contain the nessesary methods.
        """
        
        if not hasattr(self, "check_click") or not hasattr(self, "check_hover"):
            raise Exception("Base Button class should not be used by itself. Use another class like RectButton instead")

        if self.check_click():
            self.call_func(self.on_click)
        elif self.check_hover():
            self.call_func(self.on_hover)
        else:
            self.call_func(self.on_normal)

        self.draw()


class RectButton(Button):
    """
    A rectangular button.

    This inherits from the Button class.

    Attributes
    ----------
    This class inherits from Button, so contains all the attributes that Button does.
    It also contains these additional attributes:
    background_colour : tuple[int]
        the colour of the button
    width : int
        the width (parallel to the x axis) of the button
    corner_radius : int, optional
        the radius of the circles on the rounded corners of the rectangle where -1 means sharp edges and not rounded corners (defaults to -1)
    
    Methods
    -------
    This class inherits from Button, so contains all the metehods that Button does.
    It also contains these additional methods:
    mouse_over()
        return whether the mouse is colliding with, or within the region of, the button rectangle
    draw()
        draw a rectangle with the appropriate width, height, position and colour
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], width: int, height: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, corner_radius: int = -1, click_once: bool = True) -> None:
        """Construct the necessary attributes for the RectButton object"""
        super().__init__(surface, x, y, on_click, on_hover, on_normal, click_once)

        self.width = width
        self.height = height

        self.background_colour = background_colour

        self.corner_radius = corner_radius

    def mouse_over(self) -> bool:
        """Return whether the mouse is colliding with, or within the region of, the button rectangle"""
        x, y = pygame.mouse.get_pos()

        min_x = self.x - self.width // 2
        max_x = self.x + self.width // 2
        min_y = self.y - self.height // 2
        max_y = self.y + self.height // 2

        return min_x <= x <= max_x and min_y <= y <= max_y
    
    def draw(self) -> None:
        """Draw a rectangle with the appropriate width, height, position and colour"""
        rect = pygame.Rect(self.x - self.width // 2, self.y - self.height // 2, self.width, self.height)
        pygame.draw.rect(self.surface, self.background_colour, rect, border_radius=self.corner_radius)

    get_width = lambda self: self.width


class CircleButton(Button):
    """
    A circular button.

    This inherits from the Button class.

    Attributes
    ----------
    This class inherits from Button, so contains all the attributes that Button does.
    It also contains these additional attributes:
    background_colour : tuple[int]
        the colour of the button
    radius : int
        the radius of the circle
        
    Methods
    -------
    This class inherits from Button, so contains all the metehods that Button does.
    It also contains these additional methods:
    mouse_over()
        return whether the mouse position is within the circle
    draw()
        draw a circle with the appropriate radius, position and colour
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], radius: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the nessesary attributes for the CircleButton object"""
        super().__init__(surface, x, y, on_click, on_hover, on_normal, click_once)

        self.background_colour = background_colour

        self.radius = radius
        self.radius_sq = radius**2

    def mouse_over(self) -> bool:
        """Return whether the mouse position is within the circle"""
        x, y = pygame.mouse.get_pos()

        dist_sq = (x - self.x)**2 + (y - self.y)**2

        return dist_sq < self.radius_sq
    
    def draw(self) -> None:
        """Draw a circle with the appropriate radius, position and colour"""
        center = self.get_pos()
        pygame.draw.circle(self.surface, self.background_colour, center, self.radius)

    get_width = lambda self: self.radius * 2


class PolygonButton(Button):
    """
    A button in the shape of any convex polygon.
    Convex polygons are **not** supported and will result in unexpected behaviour.

    This inherits from the Button class.

    Attributes
    ----------
    This class inherits from Button, so contains all the attributes that Button does.
    It also contains these additional attributes:
    background_colour: tuple[int]
        the colour of the button
    points : list[tuple[int]]
        a list of x, y coordinates that correspond to each point on the polygon
        
    Methods
    -------
    This class inherits from Button, so contains all the metehods that Button does.
    It also contains these additional methods:
    get_center(points: list[tuple[int]])
        return the center of the polygon by averaging the position of each point
    get_inequalities(points: list[tuple[int]], center_x: int, center_y: int)
        return a list of inequalities that describe the button region
    mouse_over()
        return whether the mouse position is within the polygon using inequalities
    draw()
        draw a polygon with the appropriate points and colour
    """

    def __init__(self, surface: pygame.Surface, points: list[tuple[int]], background_colour: tuple[int], on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the necessary attributes for the PolygonButton object"""
        x, y = self.get_center(points)
        
        super().__init__(surface, x, y, on_click, on_hover, on_normal, click_once)

        self.points = points
        self.background_colour = background_colour

        self.inequalities = self.get_inequalities(points, x, y)

    def get_center(self, points: list[tuple[int]]) -> tuple[int]:
        """
        Return the center of the polygon by averaging the position of each point.

        Parameters
        ----------
        points : list[tuple[int]]
            a list of x, y coordinates that correspond to each point on the polygon

        Returns
        -------
        tuple[int]
            x, y coordinates to describe the center of the polygon
        """
        x_values = [i[0] for i in points]
        y_values = [i[1] for i in points]

        x = sum(x_values) // len(x_values)
        y = sum(y_values) // len(y_values)

        return x, y
    
    def get_inequalities(self, points: list[tuple[int]], center_x: int, center_y: int) -> list[tuple[float, float, bool]]:
        """
        Return a list of inequalities that describe the button region.

        Parameters
        ----------
        points : list[tuple[int]]
            a list of x, y coordinates that correspond to each point on the polygon
        center_x : int
            the x coordinate of the center of the polygon
        center_y : int
            the y coordinate of the center of the polygon

        Returns
        -------
        list[tuple[float, float, bool]]
            a list of linear inequalities in the form: line_gradient, y_intercept, greater_than (each line is interpreted as having the equation y=mx + c)
        """
        inequalities = []
        for i, x in enumerate(points):
            try:
                m = (x[1] - points[i - 1][1]) / (x[0] - points[i - 1][0])
            except ZeroDivisionError:
                m = 1e10

            c = x[1] - m * x[0]
            greater_than = (m * center_x + c) < center_y

            inequalities.append((m, c, greater_than))

        return inequalities

    def mouse_over(self) -> None:
        """Return whether the mouse position is within the polygon using inequalities."""
        x, y = pygame.mouse.get_pos()

        for m, c, greater_than in self.inequalities:
            intersect_y = m * x + c
            actual_greater_than = intersect_y < y

            if actual_greater_than != greater_than:
                return False
            
        return True
    
    def draw(self) -> None:
        """Draw a polygon with the appropriate points and colour."""
        pygame.draw.polygon(self.surface, self.background_colour, self.points)


class ImageButton(RectButton):
    """
    A button as an image.

    This inherits from the RectButton class.

    Attributes
    ----------
    This class inherits from RectButton, so contains all the attributes that RectButton does.
    It also contains these additional attributes:
    image_path : str
        the file path to the desired image
    image : pygame.Surface
        the image of the button in a form useable for pygame

    Methods
    -------
    This class inherits from RectButton, so contains all the methods that RectButton does.
    It also contains these additional methods:
    update_image()
        update the image attribute of the object

    The following methods are overwritten:
    draw()
        blit the image to the screen
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, width: int, height: int, image_path: str, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the necessary attributes for the ImageButton object"""
        super().__init__(surface, x, y, None, width, height, on_click, on_hover, on_normal, -1, click_once)

        self.update_image(image_path)

    def update_image(self, new_image_path: str) -> None:
        """Update the image attribute of the object."""
        image = pygame.image.load(new_image_path)
        scaled_image = pygame.transform.scale(image, (self.width, self.height))

        self.image = scaled_image

    def draw(self) -> None:
        """Blit the image to the screen."""
        image_x = self.x - self.width // 2
        image_y = self.y - self.height // 2

        self.surface.blit(self.image, (image_x, image_y))


class BorderedRectButton(RectButton):
    """
    A rectangluar button with a border.

    This inherits from the RectButton class.

    Attributes
    ----------
    This class inherits from RectButton, so contains all the attributes that RectButton does.
    It also contains these additional attributes:
    border_colour : tuple[int]
        the colour of the border
    border_width : int
        the width, or thickness, of the border

    Methods
    -------
    This class inherits from RectButton, so contains all the methods that RectButton does.
    The following methods are overwritten:
    draw()
        draw a rectangle and border
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], width: int, height: int, border_width: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, corner_radius: int = -1, click_once: bool = True) -> None:
        """Construct the necessary attributes for the BorderedRectButton object."""
        super().__init__(surface, x, y, background_colour, width, height, on_click, on_hover, on_normal, corner_radius, click_once)

        self.border_colour = border_colour
        self.border_width = border_width

    def draw(self) -> None:
        """Draw a rectangle and border."""
        rect = pygame.Rect(self.x - self.width // 2, self.y - self.height // 2, self.width, self.height)

        pygame.draw.rect(self.surface, self.background_colour, rect, border_radius=self.corner_radius)
        pygame.draw.rect(self.surface, self.border_colour, rect, self.border_width, self.corner_radius)


class BorderedCircleButton(CircleButton):
    """
    A circular button with a border.

    This inherits from the CircleButton class.

    Attributes
    ----------
    This class inherits from CircleButton, so contains all the attributes that CircleButton does.
    It also contains these additional attributes:
    border_colour : tuple[int]
        the colour of the border
    border_width : int
        the width, or thickness, of the border

    Methods
    -------
    This class inherits from CircleButton, so contains all the methods that CircleButton does.
    The following methods are overwritten:
    draw()
        draw a circle and border
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], radius: int, border_width: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the necessary attributes for the BorderedCircleButton object."""
        super().__init__(surface, x, y, background_colour, radius, on_click, on_hover, on_normal, click_once)

        self.border_colour = border_colour
        self.border_width = border_width

    def draw(self) -> None:
        """Draw a circle and border."""
        pygame.draw.circle(self.surface, self.background_colour, self.get_pos(), self.radius)
        pygame.draw.circle(self.surface, self.border_colour, self.get_pos(), self.radius, self.border_width)


class BorderedPolygonButton(PolygonButton):
    """
    A polygon button with a border.

    This inherits from the PolygonButton class.

    Attributes
    ----------
    This class inherits from PolygonButton, so contains all the attributes that PolygonButton does.
    It also contains these additional attributes:
    border_colour : tuple[int]
        the colour of the border
    border_width : int
        the width, or thickness, of the border

    Methods
    -------
    This class inherits from PolygonButton, so contains all the methods that PolygonButton does.
    The following methods are overwritten:
    draw()
        draw a polygon and border
    """
    def __init__(self, surface: pygame.Surface, points: list[tuple[int]], background_colour: tuple[int], border_colour: tuple[int], border_width: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the necessary attributes for the BorderedPolygonButton object."""
        super().__init__(surface, points, background_colour, on_click, on_hover, on_normal, click_once)

        self.border_colour = border_colour
        self.border_width = border_width

    def draw(self) -> None:
        """Draw a polygon and border."""
        pygame.draw.polygon(self.surface, self.background_colour, self.points)
        pygame.draw.polygon(self.surface, self.border_colour, self.points, self.border_width)


class TextWrapper:
    """
    A wrapper object that allows text to be drawn on top of button objects.

    All buttons other than PolygonButton and BorderedPolygonButton are supported.

    Attributes
    ----------
    button_object : RectButton | CircleButton | BorderedCircleButton | BorderedRectButton
        the button object that text will be drawn upon
    text : str
        the text displayed
    font_colour : tuple[int]
        the colour of the displayed text
    font : pygame.font.Font
        the font object used to blit text
    font_name : str | None, optional
        the name of the font used (defaults to None)
    antialias : bool, optional
        whether the text is drawn with antialias (defaults to False)
    text_surface : pygame.Surface
        the surface used to render text
    text_rect : pygame.Rect
        the rect position that the text is drawn to

    Methods
    -------
    get_text()
        return a text surface and rect object to blit text
    blit_text()
        draw text to the screen
    update_text(new_text: str | None = None, new_font_colour: tuple[int] | None = None, new_font_size: int | None = None, new_font_name: str | None = None)
        change the text, font_colour, font attributes of the object
    update()
        update the TextWrapper object
    """

    def __init__(self, button_object: RectButton | CircleButton | BorderedCircleButton | BorderedRectButton, text: str, font_colour: tuple[int], font_size: int, font_name: str | None = None, antialias: bool = False) -> None:
        """Construct necessary attributes for TextWrapper object."""
        self.button_object = button_object
        
        self.text = text
        self.font_colour = font_colour
        self.font = pygame.font.Font(font_name, font_size)

        self.antialias = antialias

        self.text_surface, self.text_rect = self.get_text()

    def get_text(self) -> tuple[pygame.Surface, pygame.Rect]:
        """Return a text surface and rect object to blit text."""
        text_surf = self.font.render(self.text, self.antialias, self.font_colour, self.button_object.background_colour)
        text_rect = text_surf.get_rect()
        text_rect.center = self.button_object.get_pos()

        return text_surf, text_rect
    
    def blit_text(self) -> None:
        """Draw text to the screen."""
        self.button_object.surface.blit(self.text_surface, self.text_rect)

    def update_text(self, new_text: str, new_font_colour: tuple[int], new_font_size: int, new_font_name: str | None) -> None:
        """Change the text, font_colour, font attributes of the object."""
        utils.update_font_attrs(self, new_text, new_font_colour, new_font_name, new_font_size)

        self.text_surface, self.text_rect = self.get_text()

    def draw(self) -> None:
        """Draw button object and blit text on top."""
        self.button_object.update()
        self.blit_text()
        
    def update(self) -> None:
        """
        Update the text wrapper button.
        
        This should be called once per frame.
        """

        self.draw()


from pygame_ui_toolkit.elements import button
from pygame_ui_toolkit import utils
from pygame_ui_toolkit import pygame


class TextInput:
    """
    The base class for all text inputs.

    This class can be used on its own, although the input_button object must be made manually.

    Attributes
    ----------
    font_colour : tuple[int]
        the colour of the displayed text
    font_name : str
        the name of, or path to, the font of the displayed text
    font_size : int
        the size of the displayed text
    og_font_size : int
        the origional font size, before it is changed in the update_font_size() method
    min_font_size : int
        the minimum font size
    on_selected : callable | None, optional
        the function called every frame when the input is selected, self is passed in as the first argument (defaults to None)
    on_deselect : callable | None, optional
        the function called every frame when the input is not selected, self is passed in as the first argument (defaults to None)
    on_text_input : callable | None, optional
        the fucntion called when text is inputted into the input. The inputted text is passed into the first argument and self is passed into the second argument (defaults to None)
    text : str
        the text that has been inputted into the input
    prefix_text : str, optional
        text displayed before the inputted text on the input (defaults to "")
    antialias : bool, optional
        whether the text is drawn with antialias (defaults to False)
    change_font_size : bool, optional
        whether the font size is automatically updated when text gets too wide (defaults to True)
    input_button : button.TextWrapper
        the button text wrapper object that handles detecting events such as clicks
    normal_button_on_click : callable | None
        the input_button on_click attribute before it is changed
    selected : bool
        whether the text input is currently selected

    Methods
    -------
    create_text_wrapper(input_button: object)
        return a button.TextWrapper object
    setup_button()
        change the input button on_click attribute
    on_click()
        set selected to True and call necessary functions
    check_deselect()
        check whether mouse is outside of button and is clicking
    take_input(pygame_event_loop: pygame.event.Event)
        loop through event loop and append any key presses to self.text
    text_too_large()
        return whether the text overfits the input button
    update_font_size()
        shrink text until it fits the input button
    draw()
        draw the input button and text to the screen
    update()
        update the TextInput object
    """

    def __init__(self, input_button: object, font_colour: tuple[int], font_size: int, font_name: str | None = None, on_selected: callable = None, on_deselect: callable = None, on_text_input: callable = None, start_text: str = "", prefix_text: str = "", min_font_size: int = 10, change_font_size: bool = True, antialias: bool = False) -> None:
        """Construct the necessary attributes for the TextWrapper object."""
        self.font_colour = font_colour
        self.font_name = font_name

        self.font_size = font_size
        self.og_font_size = font_size
        self.min_font_size = min_font_size

        self.on_selected = on_selected
        self.on_deselect = on_deselect
        self.on_text_input = on_text_input

        self.text = start_text
        self.prefix_text = prefix_text

        self.antialias = antialias
        self.change_font_size = change_font_size

        self.input_button = self.create_text_wrapper(input_button)
        self.normal_button_on_click = self.input_button.button_object.on_click

        self.setup_button()

        self.selected = False

    def create_text_wrapper(self, input_button: object) -> button.TextWrapper:
        """Return a button.TextWrapper object."""
        if type(input_button) == button.TextWrapper:
            return input_button
        else:
            text_wrapper = button.TextWrapper(input_button, self.text, self.font_colour, self.font_size, self.font_name, self.antialias)

            return text_wrapper
        
    def setup_button(self) -> None:
        """Change the input button on_click attribute."""
        self.input_button.button_object.on_click = self.on_click
        
    def on_click(self) -> None:
        """Set selected to True and call necessary functions."""
        self.input_button.button_object.call_func(self.normal_button_on_click)
        self.selected = True
        utils.call_func(self.on_selected, self)

    def check_deselect(self) -> None:
        """Check whether mouse is outside of button and is clicking."""
        if pygame.mouse.get_pressed()[0] and not self.input_button.button_object.mouse_over():
            self.selected = False
            utils.call_func(self.on_deselect, self)

    def take_input(self, pygame_event_loop: list[pygame.event.Event]) -> None:
        """Loop through event loop and append any key presses to self.text."""     
        for event in pygame_event_loop:
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_BACKSPACE:
                    self.text = self.text[:-1]
                else:
                    self.text += event.unicode

                utils.call_func(self.on_text_input, self.text, self)

    def text_too_large(self) -> bool:
        """Return whether the text overfits the input button."""
        text_rect = self.input_button.get_text()[1]
        text_width = text_rect.width
        btn_width = self.input_button.button_object.get_width()

        return text_width > btn_width

    def update_font_size(self) -> None:
        """Shrink text until it fits the input button."""
        self.font_size = self.og_font_size
        self.input_button.font = pygame.font.Font(self.font_name, self.font_size)

        while self.text_too_large() and self.font_size > self.min_font_size:
            self.font_size -= 1
            self.input_button.font = pygame.font.Font(self.font_name, self.font_size)

    def draw(self) -> None:
        """Draw the input button and text to the screen."""
        self.input_button.update_text(f"{self.prefix_text}{self.text}", self.font_colour, self.font_size, self.font_name)
        self.input_button.update()

    def update(self, pygame_event_loop: list[pygame.event.Event]) -> None:
        """
        Update the TextInput object.

        This should be called once per frame.
        """
        self.check_deselect()

        if self.change_font_size:
            self.update_font_size()

        if self.selected:
            self.take_input(pygame_event_loop)

        self.draw()


class RectTextInput(TextInput):
    """
    A text input in the shape of a rectangle.

    Inherits from TextInput

    Attributes
    ----------
    all attributes from TextInput

    Methods
    -------
    all methods from TextInput
    create_button(surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], width: int, height: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int, click_once: bool)
        return a button object with the appropriate shape
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int],  width: int, height: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, on_selected: callable = None, on_deselect: callable = None, on_text_input: callable = None, corner_radius: int = -1, click_once: bool = True, text: str = "", prefix_text: str = "", min_font_size: int = 10, change_font_size: bool = True, antialias: bool = False) -> None:
        """Construct the necessary attributes for the RectTextInput object."""
        input_button = self.create_button(surface, x, y, background_colour, width, height, on_click, on_hover, on_normal, corner_radius, click_once)
        
        super().__init__(input_button, font_colour, font_size, font_name, on_selected, on_deselect, on_text_input, text, prefix_text, min_font_size, change_font_size, antialias)

    def create_button(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], width: int, height: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int, click_once: bool) -> button.RectButton:
        """Return a button object with the appropriate shape."""
        btn = button.RectButton(surface, x, y, background_colour, width, height, on_click, on_hover, on_normal, corner_radius, click_once)

        return btn
    

class CircleTextInput(TextInput):
    """
    A text input in the shape of a circle.

    Inherits from TextInput.

    Attributes
    ----------
    all attributes from TextInput

    Methods
    -------
    all methods from TextInput
    create_button(surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], width: int, height: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int, click_once: bool)
        return a button object with the appropriate shape
    """
    
    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int],  radius: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, on_selected: callable = None, on_deselect: callable = None, on_text_input: callable = None, click_once: bool = True, text: str = "", prefix_text: str = "", min_font_size: int = 10, change_font_size: bool = True, antialias: bool = False) -> None:
        """Construct the necessary attributes for the CircleTextInput object."""
        input_button = self.create_button(surface, x, y, background_colour, radius, on_click, on_hover, on_normal, click_once)
        
        super().__init__(input_button, font_colour, font_size, font_name, on_selected, on_deselect, on_text_input, text, prefix_text, min_font_size, change_font_size, antialias)

    def create_button(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], radius: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> button.CircleButton:
        """Return a button object with the appropriate shape."""
        btn = button.CircleButton(surface, x, y, background_colour, radius, on_click, on_hover, on_normal, click_once)

        return btn
    

class PolygonTextInput(TextInput):
    """
    A text input in the shape of any convex polygon.
    Concave polygons are not supported and will result in unexpected behaviour.

    Inherits from TextInput.

    Attributes
    ----------
    all attributes from TextInput

    Methods
    -------
    all methods from TextInput
    create_button(surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], width: int, height: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int, click_once: bool)
        return a button object with the appropriate shape
    """
    
    def __init__(self, surface: pygame.Surface, points: list[tuple[int]], background_colour: tuple[int], font_colour: tuple[int], font_size: int, font_name: str | None = None, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, on_selected: callable = None, on_deselect: callable = None, on_text_input: callable = None, click_once: bool = True, text: str = "", prefix_text: str = "", min_font_size: int = 10, antialias: bool = False) -> None:
        """Construct the necessary attributes for the PolygonTextInput object."""
        input_button = self.create_button(surface, points, background_colour, on_click, on_hover, on_normal, click_once)
        
        super().__init__(input_button, font_colour, font_size, font_name, on_selected, on_deselect, on_text_input, text, prefix_text, min_font_size, False, antialias)

    def create_button(self, surface: pygame.Surface, points: list[tuple[int]], background_colour: tuple[int], on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> button.PolygonButton:
        """Return a button object with the appropriate shape."""
        btn = button.PolygonButton(surface, points, background_colour, on_click, on_hover, on_normal, click_once)

        return btn


class BorderedRectTextInput(TextInput):
    """
    A text input in the shape of a rectangle with a border.

    Inherits from TextInput.

    Attributes
    ----------
    all attributes from TextInput

    Methods
    -------
    all methods from TextInput
    create_button(surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], width: int, height: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int, click_once: bool)
        return a button object with the appropriate shape
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], width: int, height: int, border_width: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, on_selected: callable = None, on_deselect: callable = None, on_text_input: callable = None, corner_radius: int = -1, click_once: bool = True, text: str = "", prefix_text: str = "", min_font_size: int = 10, change_font_size: bool = True, antialias: bool = False) -> None:
        """Construct the necessary attributes for the BorderedRectTextInput object."""
        input_button = self.create_button(surface, x, y, background_colour, border_colour, width, height, border_width, on_click, on_hover, on_normal, corner_radius, click_once)
        
        super().__init__(input_button, font_colour, font_size, font_name, on_selected, on_deselect, on_text_input, text, prefix_text, min_font_size, change_font_size, antialias)

    def create_button(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], width: int, height: int, border_width: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, corner_radius: int = -1, click_once: bool = True) -> button.BorderedRectButton:
        """Return a button object with the appropriate shape."""
        btn = button.BorderedRectButton(surface, x, y, background_colour, border_colour, width, height, border_width, on_click, on_hover, on_normal, corner_radius, click_once)

        return btn


class BorderedCircleTextInput(TextInput):
    """
    A text input in the shape of a circle with a border.

    Inherits from TextInput.

    Attributes
    ----------
    all attributes from TextInput

    Methods
    -------
    all methods from TextInput
    create_button(surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], width: int, height: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int, click_once: bool)
        return a button object with the appropriate shape
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int],  border_colour: tuple[int], radius: int, border_width: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, on_selected: callable = None, on_deselect: callable = None, on_text_input: callable = None, click_once: bool = True, text: str = "", prefix_text: str = "", min_font_size: int = 10, change_font_size: bool = True, antialias: bool = False) -> None:
        """Construct the necessary attributes for the BorderedCircleTextInput object."""
        input_button = self.create_button(surface, x, y, background_colour, border_colour, radius, border_width, on_click, on_hover, on_normal, click_once)
        
        super().__init__(input_button, font_colour, font_size, font_name, on_selected, on_deselect, on_text_input, text, prefix_text, min_font_size, change_font_size, antialias)

    def create_button(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], radius: int, border_width: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> button.BorderedCircleButton:
        """Return a button object with the appropriate shape."""
        btn = button.BorderedCircleButton(surface, x, y, background_colour, border_colour, radius, border_width, on_click, on_hover, on_normal, click_once)

        return btn
    

class BorderedPolygonTextInput(TextInput):
    """
    A text input in the shape of any convex polygon with a border.
    Concave polygons are not supported and will result in unexpected behaviour.

    Inherits from TextInput.

    Attributes
    ----------
    all attributes from TextInput

    Methods
    -------
    all methods from TextInput
    create_button(surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], width: int, height: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int, click_once: bool)
        return a button object with the appropriate shape
    """

    def __init__(self, surface: pygame.Surface, points: list[tuple[int]], background_colour: tuple[int], border_colour: tuple[int], border_width: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, on_selected: callable = None, on_deselect: callable = None, on_text_input: callable = None, click_once: bool = True, text: str = "", prefix_text: str = "", min_font_size: int = 10, antialias: bool = False) -> None:
        """Construct the necessary attributes for the BorderedPolygonTextInput object."""
        input_button = self.create_button(surface, points, background_colour, border_colour, border_width, on_click, on_hover, on_normal, click_once)
        
        super().__init__(input_button, font_colour, font_size, font_name, on_selected, on_deselect, on_text_input, text, prefix_text, min_font_size, False, antialias)

    def create_button(self, surface: pygame.Surface, points: list[tuple[int]], background_colour: tuple[int], border_colour: tuple[int], border_width: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> button.BorderedPolygonButton:
        """Return a button object with the appropriate shape."""
        btn = button.BorderedPolygonButton(surface, points, background_colour, border_colour, border_width, on_click, on_hover, on_normal, click_once)

        return btn


import setuptools


DEPENDENCIES = ["pygame>=2.0.0"]


with open("README.md", "r") as file:
    FULL_DESC = file.read()


setuptools.setup(
    name="pygame-ui-toolkit",
    author="Ben Edwards",
    version="1.0.4",
    description="A package for creating UI elements in pygame",
    long_description=FULL_DESC,
    long_description_content_type="text/markdown",
    url="https://github.com/Ben-Edwards44/pygame-ui-toolkit",
    packages=setuptools.find_packages(exclude=["examples"]),
    install_requires=DEPENDENCIES,
    python_requires=">=3.6",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Games/Entertainment",
        "Intended Audience :: Developers"
    ]
)


from pygame_ui_toolkit.elements import button
from pygame_ui_toolkit import pygame


stored_variables = {}


def assign_variables(button_object: button.Button, normal_size: tuple[int], hover_size: tuple[int], click_size: tuple[int], on_normal: callable, on_hover: callable, on_click: callable) -> None:
    """Assign variables that need to be used later to a global dict with the button object as the key."""
    
    current_variables = {
        "normal_size" : normal_size,
        "hover_size" : hover_size,
        "click_size" : click_size,

        "on_normal" : on_normal,
        "on_hover" : on_hover,
        "on_click" : on_click
    }

    stored_variables[button_object] = current_variables


def update_button_size(button_object: button.Button, event_type: str) -> None:
    """
    Change the size of a button.
    
    An exception is raised if an unsupported button (e.g. PolygonButton is used).
    """
    
    variables = stored_variables[button_object]

    button_object.call_func(variables[f"on_{event_type}"])

    size = variables[f"{event_type}_size"]

    t = type(button_object)

    if t == button.RectButton or t == button.BorderedRectButton:
        button_object.width = size[0]
        button_object.height = size[1]
    elif t == button.CircleButton or t == button.BorderedCircleButton:
        button_object.radius = size
    else:
        raise Exception("Unsupported button type for a size change. Use any rect or circle buttons only.")


def on_click_func(button_object: button.Button) -> None:
    update_button_size(button_object, "click")


def on_hover_func(button_object: button.Button) -> None:
    update_button_size(button_object, "hover")


def on_normal_func(button_object: button.Button) -> None:
    update_button_size(button_object, "normal")


def create_button(normal_size: tuple[int], hover_size: tuple[int], click_size: tuple[int], surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], on_click: callable = None, on_hover: callable = None, on_normal: callable = None, corner_radius: int = -1) -> button.RectButton:
    """
    Return a new RectButton that will automatically change size when clicked or hovered.
    
    Ensure to call the button's update() method each frame.

    Notes
    -----
    normal_size, hover_size and click_size should be provided as a tuple in the form (width, height):
    - e.g. a value of (100, 50) would result in a button with width 100 and height 50
    """

    button_object = button.RectButton(surface, x, y, background_colour, normal_size[0], normal_size[1], on_click_func, on_hover_func, on_normal_func, corner_radius, False)
    
    assign_variables(button_object, normal_size, hover_size, click_size, on_normal, on_hover, on_click)

    return button_object


def change_existing_button(button_object: object, normal_size: tuple[int] | int, hover_size: tuple[int] | int, click_size: tuple[int] | int) -> None:
    """
    Update an existing button to make it automatically change size when clicked or hovered.

    PolygonButton and BorderedPolygonButton types are not supported.

    Notes
    -----
    - this will change the click_once attribute to False
    - if button_object is a RectButton or BorderedRectButton then normal_size, hover_size and click_size should be a tuple in the form (width, height)
    - if button_object is a CircleButton or BorderedCircleButton then normal_size, hover_size and click_size should be a single int corresponding to the radius
    """
    
    on_click = button_object.on_click
    on_hover = button_object.on_hover
    on_normal = button_object.on_normal
    
    button_object.on_click = on_click_func
    button_object.on_hover = on_hover_func
    button_object.on_normal = on_normal_func

    button_object.click_once = False

    assign_variables(button_object, normal_size, hover_size, click_size, on_normal, on_hover, on_click)


from pygame_ui_toolkit.elements import button
from pygame_ui_toolkit import utils
from pygame_ui_toolkit import pygame


class Toggle:
    """
    The base class from which all other toggles inherit from.

    Unlike other base classes, this class can be used on its own.

    Attributes
    ----------
    button_object : object
        the button object for the toggle
    normal_button_on_click : callable | None
        the button's on_click function before it is changed by the Toggle class
    on_value_changed : callable | None, optional
        the function that is called once the value of the toggle is changed. If it accepts 0 args, the function is just called; if it accepts 1 arg, the toggle value is passed in; if it accepts 2 args, the toggle value and the toggle object are passed in (defaults to None)
    selected : bool
        whether the toggle is currently selected
    prev_selected : bool
        whether the toggle was selected on the previous frame

    Methods
    -------
    setup_button()
        prepare the button_object attribute
    on_button_click()
        toggle the selected attribute and call on_click function
    draw()
        draw the toggle to the screen
    update()
        update the Toggle object
    """

    def __init__(self, button_object: object, on_value_changed: callable = None, start_value: bool = False) -> None:
        """Construct the necessary attributes for the Toggle object."""
        self.button_object = button_object
        self.normal_button_on_click = self.button_object.on_click

        self.on_value_changed = on_value_changed

        self.selected = start_value
        self.prev_selected = start_value

        self.setup_button()
        
    def setup_button(self) -> None:
        """Prepare the button_object attribute."""
        self.button_object.click_once = True
        self.button_object.on_click = self.on_button_click

    def on_button_click(self) -> None:
        """Toggle the selected attribute and call on_click function."""
        self.button_object.call_func(self.normal_button_on_click)
        self.selected = not self.selected

    def draw(self) -> None:
        """Draw the toggle to the screen."""
        self.button_object.update()

    def update(self) -> None:
        """
        Update the Toggle object.

        This should be called once per frame.
        """
        if self.selected != self.prev_selected:
            utils.call_func(self.on_value_changed, self.selected, self)
            self.prev_selected = self.selected

        self.draw()


class TextToggle(Toggle):
    """
    A toggle with text.

    Inherits from Toggle

    Attributes
    ----------
    all attributes from Toggle
    text : str
        the text displayed on the toggle button
    text_x : int
        the x coordinate of the center of the text
    text_y : int
        the y coordinate of the center of the text
    font_colour : tuple[int]
        the colour of the displayed text
    font : pygame.Surface
        the pygame surface object used to render text
    antialias : bool
        whether the text is displayed with antialias

    Methods
    -------
    get_text()
        return a text surface and a rect object for text
    update_text(new_text: str, new_font_colour: tuple[int], new_font_size: int, new_font_name: str | None, new_text_x: int | None = None, new_text_y: int | None = None)
        change the text, font colour, font size, font name or position of the displayed text
    blit_text()
        draw text to the screen
    draw()
        draw the text and toggle to the screen
    """

    def __init__(self, button_object: object, text: str, text_x: int, text_y: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, on_value_changed: callable = None, start_value: bool = False, antialias: bool = False) -> None:
        """Construct the necessary attributes for the TextToggle object."""
        super().__init__(button_object, on_value_changed, start_value)

        self.text = text

        self.text_x = text_x
        self.text_y = text_y

        self.font_colour = font_colour
        self.font = pygame.font.Font(font_name, font_size)

        self.antialias = antialias

    def get_text(self) -> tuple[pygame.Surface, pygame.Rect]:
        """Return a text surface and a rect object for text."""
        text_surf = self.font.render(self.text, self.antialias, self.font_colour, self.button_object.background_colour)
        text_rect = text_surf.get_rect()
        text_rect.center = (self.text_x, self.text_y)

        return text_surf, text_rect
    
    def update_text(self, new_text: str, new_font_colour: tuple[int], new_font_size: int, new_font_name: str | None, new_text_x: int | None = None, new_text_y: int | None = None) -> None:
        """Change the text, font colour, font size, font name or position of the displayed text."""
        utils.update_font_attrs(self, new_text, new_font_colour, new_font_name, new_font_size)

        if new_text_x != None:
            self.text_x = new_text_x
        if new_text_y != None:
            self.text_y = new_text_y

        self.text_surface, self.text_rect = self.get_text()
    
    def blit_text(self) -> None:
        """Draw text to the screen."""
        surf, rect = self.get_text()
        self.button_object.surface.blit(surf, rect)

    def draw(self) -> None:
        """Draw the text and toggle to the screen."""
        self.button_object.update()
        self.blit_text()


class TickBox(Toggle):
    """
    A tick box without text.

    Inherits from Toggle.

    Attributes
    ----------
    all attributes from Toggle
    surface : pygame.Surface
        the surface that the toggle is drawn to
    tick_thickness : int
        the thickness, or line stroke, of the drawn tick
    tick_colour : tuple[int]
        the colour of the drawn tick

    Methods
    -------
    all methods from Toggle
    draw_tick()
        draw the tick to the screen
    draw()
        draw the tick box and - if appropriate - the tick to the screen
    """

    def __init__(self, surface: pygame.Surface, tick_thickness: int, tick_colour: tuple[int], x: int, y: int, background_colour: tuple[int], side_length: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, corner_radius: int = -1, on_value_changed: callable = None, start_value: bool = False) -> None:
        """Construct the necessary attributes for the TickBox object."""
        button_object = button.RectButton(surface, x, y, background_colour, side_length, side_length, on_click, on_hover, on_normal, corner_radius, True)
        
        self.surface = surface

        self.tick_thickness = tick_thickness
        self.tick_colour = tick_colour

        super().__init__(button_object, on_value_changed, start_value)
    
    def draw_tick(self) -> None:
        """Draw the tick to the screen."""
        width = self.button_object.width
        height = self.button_object.height

        x = self.button_object.x - width // 2
        y = self.button_object.y - height // 2

        left = (x + int(width * 0.2), y + int(height * 0.6))
        middle = (x + int(width * 0.4), y + int(height * 0.9))
        right = (x + int(width * 0.9), y + int(height * 0.2))

        pygame.draw.line(self.surface, self.tick_colour, left, middle, self.tick_thickness)
        pygame.draw.line(self.surface, self.tick_colour, middle, right, self.tick_thickness)

    def draw(self) -> None:
        """Draw the tick box and - if appropriate - the tick to the screen."""
        self.button_object.update()

        if self.selected:
            self.draw_tick()


class TickBoxToggle:
    """
    A tick box toggle with text.

    Attributes
    ----------
    surface : pygame.Surface
        the surface that the toggle is drawn upon
    x : int
        the x coordinate of the toggle
    y : int
        the y coordinate of the toggle
    width : int
        the width of the toggle
    height : int
        the height of the toggle
    outer_box_colour : tuple[int]
        the colour of the outer box (the text box)
    outer_box_corner_radius : int, optional
        the radius of the rounded corners of the outer box, where -1 means no rounded corners (defaults to -1)
    text_to_right : bool, optional
        whether the text is drawn to the right of the tick box (defaults to True)
    text : str
        the displayed text
    font_colour : tuple[int]
        the colour of the displayed text
    font : pygame.Surface
        the surface object used to render text
    tick_box : TickBox
        the tick box object for the toggle
    text_surface : pygame.Surface
        the surface object used to render text
    text_rect : pygame.Rect
        the rect position that the text is drawn to
    selected : bool
        whether the toggle is currrently selected

    Methods
    -------
    create_tick_box(tick_thickness: int, tick_colour: tuple[int], background_colour: tuple[int], dist_from_edge: int, height_offset: int, on_click: callable, on_hover: callable, on_normal: callable, on_value_changed: callable, corner_radius: int, start_value: bool)
        return the tick box object with the appropriate size and position
    find_text_pos()
        return the x, y position of the text
    get_text()
        return the text surface and rect objects to draw text
    update_text(new_text: str, new_font_colour: tuple[int], new_font_size: int, new_font_name: str | None)
        change the text, font colour, font size, or font name of the displayed text
    draw()
        draw the outer box, tick box and text to the screen
    update()
        update the TickBoxToggle object
    """

    def __init__(self, surface: pygame.Surface, tick_thickness: int, tick_colour: tuple[int], tick_box_colour: tuple[int], outer_box_colour: tuple[int], x: int, y: int, width: int, height: int, text: str, font_colour: tuple[int], font_size: int, font_name: str | None = None, dist_from_edge: int = 5, height_offset: int = 10, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, on_value_changed: callable = None, inner_corner_radius: int = -1, outer_corner_radius: int = -1, start_value: bool = False, text_to_right: bool = True, antialias: bool = False) -> None:
        """Construct the necessary attributes for the TickBoxToggle object."""
        self.surface = surface

        self.x = x
        self.y = y

        self.width = width
        self.height = height

        self.outer_box_colour = outer_box_colour
        self.outer_corner_radius = outer_corner_radius

        self.text_to_right = text_to_right

        self.text = text

        self.font_colour = font_colour
        self.font = pygame.font.Font(font_name, font_size)

        self.antialias = antialias

        self.tick_box = self.create_tick_box(tick_thickness, tick_colour, tick_box_colour, dist_from_edge, height_offset, on_click, on_hover, on_normal, on_value_changed, inner_corner_radius, start_value)
        self.text_surface, self.text_rect = self.get_text()

        self.selected = self.tick_box.selected

    def create_tick_box(self, tick_thickness: int, tick_colour: tuple[int], background_colour: tuple[int], dist_from_edge: int, height_offset: int, on_click: callable, on_hover: callable, on_normal: callable, on_value_changed: callable, corner_radius: int, start_value: bool) -> TickBox:
        """Return the tick box object with the appropriate size and position."""
        new_height = self.height - height_offset

        if self.text_to_right:
            new_x = (self.x - self.width // 2) + (new_height // 2) + dist_from_edge
        else:
            new_x = (self.x + self.width // 2) - (new_height // 2) - dist_from_edge
        
        return TickBox(self.surface, tick_thickness, tick_colour, new_x, self.y, background_colour, new_height, on_click, on_hover, on_normal, corner_radius, on_value_changed, start_value)
    
    def find_text_pos(self) -> tuple[int, int]: 
        """Return the x, y position of the text."""    
        dist_to_center = self.tick_box.button_object.width // 2

        if self.text_to_right:
            x1 = self.x + self.width // 2
            x2 = self.tick_box.button_object.x + dist_to_center
        else:
            x1 = self.x - self.width // 2
            x2 = self.tick_box.button_object.x - dist_to_center

        x = (x1 + x2) // 2
        y = self.tick_box.button_object.y

        return (x, y)
    
    def get_text(self) -> tuple[pygame.Surface, pygame.Rect]:
        """Return the text surface and rect objects to draw text."""
        text_surf = self.font.render(self.text, self.antialias, self.font_colour, self.outer_box_colour)
        text_rect = text_surf.get_rect()

        text_pos = self.find_text_pos()
        text_rect.center = text_pos

        return text_surf, text_rect
    
    def update_text(self, new_text: str, new_font_colour: tuple[int], new_font_size: int, new_font_name: str | None) -> None:
        """Change the text, font colour, font size, or font name of the displayed text."""
        utils.update_font_attrs(self, new_text, new_font_colour, new_font_name, new_font_size)

        self.text_surface, self.text_rect = self.get_text()

    def draw(self) -> None:
        """Draw the outer box, tick box and text to the screen."""
        outer_rect = pygame.Rect(self.x - self.width // 2, self.y - self.height // 2, self.width, self.height)
        pygame.draw.rect(self.surface, self.outer_box_colour, outer_rect, border_radius=self.outer_corner_radius)

        self.surface.blit(self.text_surface, self.text_rect)

        self.tick_box.draw()

    def update(self) -> None:
        """
        Update the TickBoxToggle object.

        This should be called once per frame.
        """

        self.tick_box.update()
        self.draw()

        self.selected = self.tick_box.selected